@@ -1,5 +1,4 @@
 import logging
-import os
 import time
 import threading
 import math
@@ -275,7 +274,7 @@ class Database(MutableMapping):
     def __init__(self, lock, *args, **kwargs):
         self._dict = dict(*args, **kwargs)
         self.time = 0.0
-        self._watches = defaultdict(set)      # key to set of connections
+        self._watches = defaultdict(weakref.WeakSet)      # key to set of connections
         self.condition = threading.Condition(lock)
         self._change_callbacks = set()
 
@@ -641,9 +640,13 @@ class FakeServer:
         self.psubscribers = defaultdict(weakref.WeakSet)
         self.lastsave = int(time.time())
         self.connected = True
+        # List of weakrefs to sockets that are being closed lazily
+        self.closed_sockets = []
 
 
 class FakeSocket:
+    _connection_error_class = redis.ConnectionError
+
     def __init__(self, server):
         self._server = server
         self._db = server.dbs[0]
@@ -665,7 +668,12 @@ class FakeSocket:
         self._parser.send(None)
 
     def put_response(self, msg):
-        self.responses.put(msg)
+        # redis.Connection.__del__ might call self.close at any time, which
+        # will set self.responses to None. We assume this will happen
+        # atomically, and the code below then protects us against this.
+        responses = self.responses
+        if responses:
+            responses.put(msg)
 
     def pause(self):
         self._paused = True
@@ -683,13 +691,24 @@ class FakeSocket:
         # `FakeSelector` before it is ever used.
         return 0
 
+    def _cleanup(self, server):
+        """Remove all the references to `self` from `server`.
+
+        This is called with the server lock held, but it may be some time after
+        self.close.
+        """
+        for subs in server.subscribers.values():
+            subs.discard(self)
+        for subs in server.psubscribers.values():
+            subs.discard(self)
+        self._clear_watches()
+
     def close(self):
-        with self._server.lock:
-            for subs in self._server.subscribers.values():
-                subs.discard(self)
-            for subs in self._server.psubscribers.values():
-                subs.discard(self)
-            self._clear_watches()
+        # Mark ourselves for cleanup. This might be called from
+        # redis.Connection.__del__, which the garbage collection could call
+        # at any time, and hence we can't safely take the server lock.
+        # We rely on list.append being atomic.
+        self._server.closed_sockets.append(weakref.ref(self))
         self._server = None
         self._db = None
         self.responses = None
@@ -807,7 +826,7 @@ class FakeSocket:
 
     def sendall(self, data):
         if not self._server.connected:
-            raise redis.ConnectionError(CONNECTION_ERROR_MSG)
+            raise self._connection_error_class(CONNECTION_ERROR_MSG)
         if isinstance(data, str):
             data = data.encode('ascii')
         self._parser.send(data)
@@ -815,10 +834,21 @@ class FakeSocket:
     def _process_command(self, fields):
         if not fields:
             return
+        func_name = None
         try:
             func, func_name = self._name_to_func(fields[0])
             sig = func._fakeredis_sig
             with self._server.lock:
+                # Clean out old connections
+                while True:
+                    try:
+                        weak_sock = self._server.closed_sockets.pop()
+                    except IndexError:
+                        break
+                    else:
+                        sock = weak_sock()
+                        if sock:
+                            sock._cleanup(self._server)
                 now = time.time()
                 for db in self._server.dbs.values():
                     db.time = now
@@ -835,6 +865,11 @@ class FakeSocket:
                 # TODO: should not apply if the exception is from _run_command
                 # e.g. watch inside multi
                 self._transaction_failed = True
+            if func_name == 'exec' and exc.value.startswith('ERR '):
+                exc.value = 'EXECABORT Transaction discarded because of: ' + exc.value[4:]
+                self._transaction = None
+                self._transaction_failed = False
+                self._clear_watches()
             result = exc
         result = self._decode_result(result)
         if not isinstance(result, NoResponse):
@@ -1287,6 +1322,7 @@ class FakeSocket:
             raise SimpleError(WITHOUT_MULTI_MSG.format('EXEC'))
         if self._transaction_failed:
             self._transaction = None
+            self._clear_watches()
             raise SimpleError(EXECABORT_MSG)
         transaction = self._transaction
         self._transaction = None
@@ -2648,39 +2684,9 @@ class FakeSelector(BaseSelector):
 class FakeConnection(redis.Connection):
     description_format = "FakeConnection<db=%(db)s>"
 
-    def __init__(self, server, db=0, username=None, password=None,
-                 socket_timeout=None, socket_connect_timeout=None,
-                 socket_keepalive=False, socket_keepalive_options=None,
-                 socket_type=0, retry_on_timeout=False,
-                 encoding='utf-8', encoding_errors='strict',
-                 decode_responses=False, parser_class=_DummyParser,
-                 socket_read_size=65536, health_check_interval=0,
-                 client_name=None):
-        self.pid = os.getpid()
-        self.db = db
-        self.username = username
-        self.client_name = client_name
-        self.password = password
-        # Allow socket attributes to be passed in and saved even if they aren't used
-        self.socket_timeout = socket_timeout
-        self.socket_connect_timeout = socket_connect_timeout or socket_timeout
-        self.socket_keepalive = socket_keepalive
-        self.socket_keepalive_options = socket_keepalive_options or {}
-        self.socket_type = socket_type
-        self.retry_on_timeout = retry_on_timeout
-        self.encoder = redis.connection.Encoder(encoding, encoding_errors, decode_responses)
-        self._description_args = {'db': self.db}
-        self._connect_callbacks = []
-        self._buffer_cutoff = 6000
-        self._server = server
-        # self._parser isn't used for anything, but some of the
-        # base class methods depend on it and it's easier not to
-        # override them.
-        self._parser = parser_class(socket_read_size=socket_read_size)
-        self._sock = None
-        # added in redis==3.3.0
-        self.health_check_interval = health_check_interval
-        self.next_health_check = 0
+    def __init__(self, *args, **kwargs):
+        self._server = kwargs.pop('server')
+        super().__init__(*args, **kwargs)
 
     def connect(self):
         super().connect()
@@ -2798,9 +2804,9 @@ class FakeRedisMixin:
         pool = self.connection_pool
         pool.connection_class = FakeConnection
         pool.connection_kwargs['server'] = server
-        for key in ['password', 'host', 'port', 'path']:
-            if key in pool.connection_kwargs:
-                del pool.connection_kwargs[key]
+        # FakeConnection cannot handle the path kwarg (present when from_url
+        # is called with a unix socket)
+        pool.connection_kwargs.pop('path', None)
         return self
 
 
